import logging
from pathlib import Path
from typing import Union

import pandas as pd
import spectrum_fundamentals.constants as c
from spectrum_fundamentals.mod_string import internal_without_mods, maxquant_to_internal

from .search_results import SearchResults, filter_valid_prosit_sequences

logger = logging.getLogger(__name__)


class MaxQuant(SearchResults):
    """Handle search results from MaxQuant."""

    @staticmethod
    def add_tmt_mod(mass: float, seq: str, unimod_tag: str) -> float:
        """
        Add tmt modification.

        :param mass: mass without tmt modification
        :param seq: sequence of the peptide
        :param unimod_tag: UNIMOD tag for the modification
        :return: mass as float
        """
        num_of_tmt = seq.count(unimod_tag)
        mass += num_of_tmt * c.MOD_MASSES[f"{unimod_tag}"]
        return mass

    @staticmethod
    def read_result(path: Union[str, Path], tmt_labeled: str) -> pd.DataFrame:
        """
        Function to read a msms txt and perform some basic formatting.

        :param path: path to msms.txt to read
        :param tmt_labeled: tmt label as str
        :return: pd.DataFrame with the formatted data
        """
        logger.info("Reading msms.txt file")
        df = pd.read_csv(
            path,
            usecols=lambda x: x.upper()
            in [
                "RAW FILE",
                "SCAN NUMBER",
                "MODIFIED SEQUENCE",
                "CHARGE",
                "SCAN EVENT NUMBER",
                "LABELING STATE",
                "MASS",  # = Calculated Precursor mass; TODO get column with experimental Precursor mass instead
                "SCORE",
                "REVERSE",
            ],
            sep="\t",
        )
        logger.info("Finished reading msms.txt file")

        # Standardize column names
        df.columns = df.columns.str.upper()
        df.columns = df.columns.str.replace(" ", "_")

        df = MaxQuant.update_columns_for_prosit(df, tmt_labeled)
        return filter_valid_prosit_sequences(df)

    @staticmethod
    def update_columns_for_prosit(df: pd.DataFrame, tmt_labeled: str) -> pd.DataFrame:
        """
        Update columns of df to work with Prosit.

        :param df: df to modify
        :param tmt_labeled: True if tmt labeled
        :return: modified df as pd.DataFrame
        """
        df.rename(columns={"CHARGE": "PRECURSOR_CHARGE"}, inplace=True)

        df["REVERSE"].fillna(False, inplace=True)
        df["REVERSE"].replace("+", True, inplace=True)
        logger.info("Converting MaxQuant peptide sequence to internal format")
        if tmt_labeled != "":
            unimod_tag = c.TMT_MODS[tmt_labeled]
            logger.info("Adding TMT fixed modifications")
            df["MODIFIED_SEQUENCE"] = maxquant_to_internal(
                df["MODIFIED_SEQUENCE"].to_numpy(),
                fixed_mods={"C": "C[UNIMOD:4]", "^_": f"_{unimod_tag}-", "K": f"K{unimod_tag}"},
            )
            df["MASS"] = df.apply(lambda x: MaxQuant.add_tmt_mod(x.MASS, x.MODIFIED_SEQUENCE, unimod_tag), axis=1)
            if "msa" in tmt_labeled:
                logger.info("Replacing phospho by dehydration for Phospho-MSA")
                df["MODIFIED_SEQUENCE_MSA"] = df["MODIFIED_SEQUENCE"].str.replace(
                    "[UNIMOD:21]", "[UNIMOD:23]", regex=False
                )
        elif "LABELING_STATE" in df.columns:
            logger.info("Adding SILAC fixed modifications")
            df.loc[df["LABELING_STATE"] == 1, "MODIFIED_SEQUENCE"] = maxquant_to_internal(
                df[df["LABELING_STATE"] == 1]["MODIFIED_SEQUENCE"].to_numpy(),
                fixed_mods={"C": "C[UNIMOD:4]", "K": "K[UNIMOD:259]", "R": "R[UNIMOD:267]"},
            )
            df.loc[df["LABELING_STATE"] != 1, "MODIFIED_SEQUENCE"] = maxquant_to_internal(
                df[df["LABELING_STATE"] != 1]["MODIFIED_SEQUENCE"].to_numpy()
            )
            df["MASS"] = df.apply(lambda x: MaxQuant.add_tmt_mod(x.MASS, x.MODIFIED_SEQUENCE, "[UNIMOD:259]"), axis=1)
            df["MASS"] = df.apply(lambda x: MaxQuant.add_tmt_mod(x.MASS, x.MODIFIED_SEQUENCE, "[UNIMOD:267]"), axis=1)
            df.drop(columns=["LABELING_STATE"], inplace=True)
        else:
            df["MODIFIED_SEQUENCE"] = maxquant_to_internal(df["MODIFIED_SEQUENCE"].to_numpy())
        df["SEQUENCE"] = internal_without_mods(df["MODIFIED_SEQUENCE"])
        df["PEPTIDE_LENGTH"] = df["SEQUENCE"].apply(lambda x: len(x))

        return df



import os

import numpy as np
import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_spectronaut, internal_without_mods

from .spectral_library import SpectralLibrary


class Spectronaut(SpectralLibrary):
    """Main to initialze a Spectronaut obj."""

    # Check spectronaut folder for output format.

    def write(self):
        """
        Writing method.

        Splits intermediate dataframe into chunks of 10k lines, explodes them,
        and filters for relevant peaks before storing it in self.out_path.
        """
        n = 7000  # split df into chunks of size n
        initial = not os.path.isfile(self.out_path)
        for _, segment in self.spectra_output.groupby(np.arange(len(self.spectra_output)) // n):
            segment = segment.explode(
                ["intensities", "fragment_mz", "fragment_types", "fragment_numbers", "fragment_charges"]
            )
            segment = segment[segment["intensities"] > 0]  # set to >= if 0 should be kept
            segment.rename(
                columns={
                    "intensities": "RelativeIntensity",
                    "fragment_mz": "FragmentMz",
                    "fragment_types": "FragmentType",
                    "fragment_numbers": "FragmentNumber",
                    "fragment_charges": "FragmentCharge",
                },
                inplace=True,
            )
            segment["FragmentLossType"] = "noloss"
            if len(list(self.grpc_output)) > 2:
                segment = segment[
                    [
                        "RelativeIntensity",
                        "FragmentMz",
                        "ModifiedPeptide",
                        "LabeledPeptide",
                        "StrippedPeptide",
                        "PrecursorCharge",
                        "PrecursorMz",
                        "iRT",
                        "proteotypicity",
                        "FragmentNumber",
                        "FragmentType",
                        "FragmentCharge",
                        "FragmentLossType",
                    ]
                ]
            else:
                segment = segment[
                    [
                        "RelativeIntensity",
                        "FragmentMz",
                        "ModifiedPeptide",
                        "LabeledPeptide",
                        "StrippedPeptide",
                        "PrecursorCharge",
                        "PrecursorMz",
                        "iRT",
                        "FragmentNumber",
                        "FragmentType",
                        "FragmentCharge",
                        "FragmentLossType",
                    ]
                ]
            segment.to_csv(self.out_path, mode="a", header=initial, index=False)

    def prepare_spectrum(self):
        """Converts grpc output and metadata dataframe into spectronaut format."""
        intensities = self.grpc_output[list(self.grpc_output)[0]]["intensity"]
        fragment_mz = self.grpc_output[list(self.grpc_output)[0]]["fragmentmz"]
        annotation = self.grpc_output[list(self.grpc_output)[0]]["annotation"]
        fragment_types = annotation["type"]
        fragment_numbers = annotation["number"]
        fragment_charges = annotation["charge"]
        irt = self.grpc_output[list(self.grpc_output)[1]]
        irt = irt.flatten()
        if len(list(self.grpc_output)) > 2:
            proteotypicity = self.grpc_output[list(self.grpc_output)[2]]
            proteotypicity = proteotypicity.flatten()
        modified_sequences_spec = internal_to_spectronaut(
            self.spectra_input["MODIFIED_SEQUENCE"].apply(lambda x: "_" + x + "_")
        )
        modified_sequences = self.spectra_input["MODIFIED_SEQUENCE"]

        labelled_sequences = internal_without_mods(modified_sequences)
        stripped_peptide = internal_without_mods(modified_sequences)
        charges = self.spectra_input["PRECURSOR_CHARGE"]
        precursor_masses = self.spectra_input["MASS"]
        precursor_mz = (precursor_masses + (charges * PARTICLE_MASSES["PROTON"])) / charges

        inter_df = pd.DataFrame(
            data={
                "ModifiedPeptide": modified_sequences_spec,
                "LabeledPeptide": labelled_sequences,
                "StrippedPeptide": stripped_peptide,
                "PrecursorCharge": charges,
                "PrecursorMz": precursor_mz,
            }
        )
        inter_df["iRT"] = irt.tolist()
        if len(list(self.grpc_output)) > 2:
            inter_df["proteotypicity"] = proteotypicity.tolist()
        inter_df["intensities"], inter_df["fragment_mz"] = intensities.tolist(), fragment_mz.tolist()
        inter_df["fragment_types"] = fragment_types.tolist()
        inter_df["fragment_numbers"] = fragment_numbers.tolist()
        inter_df["fragment_charges"] = fragment_charges.tolist()

        self.spectra_output = inter_df



#!/usr/bin/env python
"""Command-line interface."""
import click
from rich import traceback


@click.command()
@click.version_option(version="0.3.3", message=click.style("spectrum_io Version: 0.3.3"))
def main() -> None:
    """spectrum_io."""


if __name__ == "__main__":
    traceback.install()
    main(prog_name="spectrum_io")  # pragma: no cover



from abc import abstractmethod
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd


class SpectralLibrary:
    """Main to initialze a SpectralLibrary obj."""

    def __init__(
        self,
        input_dataframe: pd.DataFrame,
        grpc_dict: dict,
        output_path: Union[str, Path],
        min_intensity_threshold: Optional[float] = 0.05,
        chunksize: Optional[int] = None,
    ):
        """
        Initialize a SpectralLibrary obj.

        :param input_dataframe: dataframe of sequences, charges, and masses of all library peptides
        :param grpc_dict: GRPC client output dictionary with spectrum, irt, and proteotypicity prediction
        :param output_path: path to output file including file name
        :param min_intensity_threshold: optional filter for low intensity peaks
        :param chunksize: optional chunksize for dlib
        """
        if isinstance(output_path, str):
            output_path = Path(output_path)
        self.spectra_input = input_dataframe
        self.grpc_output = grpc_dict
        self.out_path = output_path
        self.min_intensity_threshold = min_intensity_threshold
        self.chunksize = chunksize

    def load(self):
        """Load predictions from hdf5 file."""

    @abstractmethod
    def write(self):
        """Write predictions."""
        pass

    @abstractmethod
    def prepare_spectrum(self):
        """Prepare spectrum."""
        pass



"""Initialize spectral library."""
import logging

from . import digest
from .dlib import DLib
from .msp import MSP
from .spectral_library import SpectralLibrary
from .spectronaut import Spectronaut

logger = logging.getLogger(__name__)



import logging
from pathlib import Path
from typing import Union

import pandas as pd
import spectrum_fundamentals.constants as c
from pyteomics import pepxml
from spectrum_fundamentals.mod_string import internal_without_mods
from tqdm import tqdm

from .search_results import SearchResults, filter_valid_prosit_sequences

logger = logging.getLogger(__name__)


class MSFragger(SearchResults):
    """Handle search results from MSFragger."""

    @staticmethod
    def read_result(path: Union[str, Path], tmt_labeled: str) -> pd.DataFrame:
        """
        Function to read a msms txt and perform some basic formatting.

        :param path: path to pepXML folder or single pepXML file to read
        :param tmt_labeled: tmt label as str
        :raises FileNotFoundError: in case the given path is neither a file, nor a directory.
        :return: pd.DataFrame with the formatted data
        """
        if isinstance(path, str):
            path = Path(path)

        if path.is_file():
            file_list = [path]
        elif path.is_dir():
            file_list = list(path.rglob("*.pepXML"))
        else:
            raise FileNotFoundError(f"{path} could not be found.")

        ms_frag_results = []
        for pep_xml_file in tqdm(file_list):
            ms_frag_results.append(pepxml.DataFrame(str(pep_xml_file)))

        df = pd.concat(ms_frag_results)

        df = update_columns_for_prosit(df, "")
        return filter_valid_prosit_sequences(df)


def update_columns_for_prosit(df, tmt_labeled: str) -> pd.DataFrame:
    """
    Update columns of df to work with Prosit.

    :param df: df to modify
    :param tmt_labeled: True if tmt labeled
    :return: modified df as pd.DataFrame
    """
    df["REVERSE"] = df["protein"].apply(lambda x: "rev" in str(x))
    df["RAW_FILE"] = df["spectrum"].apply(lambda x: x.split(".")[0])
    df["MASS"] = df["precursor_neutral_mass"]
    df["PEPTIDE_LENGTH"] = df["peptide"].apply(lambda x: len(x))
    df["MODIFIED_SEQUENCE"] = msfragger_to_internal(df["modified_peptide"])
    df.rename(
        columns={
            "assumed_charge": "PRECURSOR_CHARGE",
            "index": "SCAN_EVENT_NUMBER",
            "peptide": "SEQUENCE",
            "start_scan": "SCAN_NUMBER",
            "hyperscore": "SCORE",
        },
        inplace=True,
    )
    df["SEQUENCE"] = internal_without_mods(df["MODIFIED_SEQUENCE"])
    return df[
        [
            "RAW_FILE",
            "SCAN_NUMBER",
            "MODIFIED_SEQUENCE",
            "PRECURSOR_CHARGE",
            "SCAN_EVENT_NUMBER",
            "MASS",
            "SCORE",
            "REVERSE",
            "SEQUENCE",
            "PEPTIDE_LENGTH",
        ]
    ]


def msfragger_to_internal(modstrings: pd.Series):
    """
    Transform modstring from msfragger format to internal format.

    This function takes a modstrings column from a pandas dataframe and converts each
    supported modification (M[147] and C[160]) to the internal representation that is
    M[UNIMOD:35] and C[UNIMOD:4], respectively. Since C is considered a fixed modification,
    every occurence of a C is transformed to C[UNIMOD:4] as well.

    :param modstrings: pd.Series containing the msfragger modstrings
    :return: pd.Series with internal modstrings
    """
    modstrings = modstrings.str.replace("M[147]", "M[UNIMOD:35]", regex=False)
    modstrings = modstrings.str.replace(r"C\[160\]|C", "C[UNIMOD:4]", regex=True)
    return modstrings



import pandas as pd
from spectrum_fundamentals.constants import PARTICLE_MASSES
from spectrum_fundamentals.mod_string import internal_to_mod_names, internal_without_mods

from .spectral_library import SpectralLibrary


class MSP(SpectralLibrary):
    """Main to initialze a MSP obj."""

    # Check msp folder for output format.
    def write(self):
        """Writing method; writes intermediate dataframe as msp format spectra."""
        out = open(self.out_path, "a")

        for _, spectrum in self.spectra_output.iterrows():
            spectrum = spectrum.to_dict()
            out.write(f"Name: {spectrum['StrippedPeptide']}/{spectrum['PrecursorCharge']}\n")
            out.write(f"MW: {spectrum['PrecursorMz']}\n")
            out.write(
                f"Comment: Parent={spectrum['PrecursorMz']} "
                f"Collision_energy={spectrum['CollisionEnergy']} "
                f"Mods={spectrum['Modifications'][0]} "
                f"ModString={spectrum['Modifications'][1]}/{spectrum['PrecursorCharge']} "
                f"iRT={spectrum['iRT']} "
            )
            if len(list(self.grpc_output)) > 2:
                out.write(f"proteotypicity={spectrum['proteotypicity']}\n")
            else:
                out.write("\n")
            out.write(f"Num peaks: {sum(elem!='N' for elem in spectrum['fragment_types'])}\n")
            for fmz, fintensity, ftype, fcharge, fnumber in zip(
                spectrum["fragment_mz"],
                spectrum["intensities"],
                spectrum["fragment_types"],
                spectrum["fragment_charges"],
                spectrum["fragment_numbers"],
            ):
                if ftype != "N":
                    fcharge = f"^{fcharge}" if fcharge != 1 else ""
                    out.write(f"{fmz}\t{fintensity}\t" f'"{ftype}{fnumber}{fcharge}/0.0ppm"\n')
        out.close()

    def prepare_spectrum(self):
        """Converts grpc output and metadata dataframe into msp format."""
        intensities = self.grpc_output[list(self.grpc_output)[0]]["intensity"]
        fragment_mz = self.grpc_output[list(self.grpc_output)[0]]["fragmentmz"]
        annotation = self.grpc_output[list(self.grpc_output)[0]]["annotation"]
        fragment_types = annotation["type"]
        fragment_numbers = annotation["number"]
        fragment_charges = annotation["charge"]
        irt = self.grpc_output[list(self.grpc_output)[1]]
        irt = irt.flatten()
        if len(list(self.grpc_output)) > 2:
            proteotypicity = self.grpc_output[list(self.grpc_output)[2]]
            proteotypicity = proteotypicity.flatten()
        modified_sequences = self.spectra_input["MODIFIED_SEQUENCE"]
        collision_energies = self.spectra_input["COLLISION_ENERGY"]

        stripped_peptide = internal_without_mods(modified_sequences)
        msp_mod_strings = internal_to_mod_names(modified_sequences)
        charges = self.spectra_input["PRECURSOR_CHARGE"]
        precursor_masses = self.spectra_input["MASS"]
        precursor_mz = (precursor_masses + (charges * PARTICLE_MASSES["PROTON"])) / charges

        inter_df = pd.DataFrame(
            data={
                "ModifiedPeptide": modified_sequences,
                "StrippedPeptide": stripped_peptide,
                "PrecursorCharge": charges,
                "PrecursorMz": precursor_mz,
                "PrecursorMass": precursor_masses,
                "CollisionEnergy": collision_energies,
                "Modifications": msp_mod_strings,
            }
        )
        inter_df["iRT"] = irt.tolist()
        if len(list(self.grpc_output)) > 2:
            inter_df["proteotypicity"] = proteotypicity.tolist()
        inter_df["intensities"], inter_df["fragment_mz"] = intensities.tolist(), fragment_mz.tolist()
        inter_df["fragment_types"] = fragment_types.tolist()
        inter_df["fragment_numbers"] = fragment_numbers.tolist()
        inter_df["fragment_charges"] = fragment_charges.tolist()

        self.spectra_output = inter_df



import logging
import threading
from pathlib import Path
from typing import List, Optional, Union

import h5py
import pandas as pd
import scipy
from scipy.sparse import coo_matrix

logger = logging.getLogger(__name__)

META_DATA_KEY = "meta_data"
INTENSITY_RAW_KEY = "raw_intensity"
INTENSITY_PRED_KEY = "pred_intensity"
MZ_RAW_KEY = "raw_mz"


def read_file(path: Union[str, Path], key: str) -> pd.DataFrame:
    """
    Read hdf5 file and return dataframe with contents.

    With possibility to partial load for memory issues.
    :param path: The path to the hdf5 file to read
    :param key: The key of the dataset/group of interest
    :return: a pandas DataFrame with contents
    """
    try:
        if key.startswith("sparse"):
            with h5py.File(path, "r") as f:
                logger.info(f"Reading sparse matrix from hdf5 file. Available keys: {f.keys()}")
                values = f[f"{key}/values"]
                i = f[f"{key}/i"]
                j = f[f"{key}/j"]
                shape = f[f"{key}/shape"]
                sparse_data = coo_matrix((values, (i, j)), shape)
                df = pd.DataFrame.sparse.from_spmatrix(sparse_data)
                if f"{key}/column_names" in f.keys():
                    df.columns = f[f"{key}/column_names"].asstr()
                if f"{key}/index" in f.keys():
                    df.index = f[f"{key}/index"]
        else:
            df = pd.read_hdf(path, key=key)
        return df
    except Exception as e:
        logger.exception(e)


def thread_this(fn):
    """Function for threading."""

    def run(*args, **kwargs):
        t = threading.Thread(target=fn, args=args, kwargs=kwargs)
        t.start()
        return t

    return run


@thread_this
def write_file(
    data_sets: List[Union[pd.DataFrame, scipy.sparse.spmatrix]],
    path: str,
    dataset_names: List[str],
    column_names: Optional[List[Optional[List[str]]]] = None,
):
    """
    Writes several datasets (spectra) to hdf5 file.

    :param data_sets: list of datasets
    :param path: path to store the file to
    :param dataset_names: list of dataset names
    :param column_names: list of column_names
    :raises TypeError: if data_set has an unexpected type
    """
    index = 0
    for data_set, dataset_name in zip(data_sets, dataset_names):
        if isinstance(data_set, pd.DataFrame):
            write_dataset(data_set, path, dataset_name)
        elif isinstance(data_set, scipy.sparse.spmatrix):
            if not isinstance(column_names, list):
                raise TypeError(f"column_names is required if data_set is of type {type(data_set)}.")
            write_dataset(data_set, path, dataset_name, mode="a", column_names=column_names[index])
            index += 1
        else:
            raise TypeError(f"data_set type not understood: {type(data_set)}.")


def write_dataset(
    data: Union[pd.DataFrame, scipy.sparse.spmatrix],
    path: str,
    dataset_name: str,
    mode: str = "w",
    compression: Optional[Union[str, bool]] = True,
    column_names: Optional[List[str]] = None,
    index: Optional[List[str]] = None,
):
    """
    Writes or appends dataset to an hdf5 file.

    :param data: The data to store. Can be a pandas DataFrame or a scipy Sparsematrix
    :param path: The path to store the file to
    :param dataset_name: The key in the hdf5 file under which to store the data
    :param mode: The method when writing the data. Use 'a' to append to an existing file or 'w' to overwrite
    :param compression: Optional, the method for compressing data. Check pandas.DataFrame.to_hdf docs for supported
            compression methods in case of providing a pandas DataFrame and h5py.Dataset docs in case of providing
            a sparse matrix. If providing False, no compression is applied; if providing True, defaults to 'zlib'
            in case of providing a pandas DataFrame and 'gzip' if providing a sparse matrix.
            standard compression depending on the data type given. Default: True
    :param column_names: Optional, additional column column_names. Ignored if providing a pandas DataFrame. Default: None
    :param index: Optional, additional index. Ignored if providing a pandas DataFrame. Default: None
    :raises AssertionError: if data_set has an unexpected type
    """
    if isinstance(compression, bool) and compression:
        if isinstance(compression, pd.DataFrame):
            compression = "zlib"
        elif isinstance(compression, scipy.sparse.spmatrix):
            compression = "gzip"
        else:
            compression = None
    try:
        if isinstance(data, pd.DataFrame):
            data.to_hdf(path, key=dataset_name, mode=mode, complib=compression)
        elif isinstance(data, scipy.sparse.spmatrix):
            with h5py.File(path, mode) as f:
                group_name = f"sparse_{dataset_name}"
                f.create_group(group_name)
                i, j, values = scipy.sparse.find(data)
                shape = data.shape
                f.create_dataset(f"{group_name}/i", data=i, compression=compression, dtype=int)
                f.create_dataset(f"{group_name}/j", data=j, compression=compression, dtype=int)
                f.create_dataset(f"{group_name}/values", data=values, compression=compression, dtype=float)
                f.create_dataset(f"{group_name}/shape", data=shape, shape=(2,), dtype=int)
                if column_names:
                    f.create_dataset(f"{group_name}/column_names", data=column_names, compression=compression)
                if index:
                    f.create_dataset(f"{group_name}/index", data=index, compression=compression)
        else:
            raise AssertionError("Only pd.DataFrame and scipy.sparse.spmatrix are supported." + type(data))
        logger.info(f"Data {'appended' if mode=='a' else 'written'} to {path}")
    except Exception as e:
        logger.exception(e)
